@@ -321,8 +321,10 @@ class PortRangeIterator(object):
         elif '-' in value:
             (start, end) = value.split('-', 1)
             if start.isdigit() and end.isdigit():
+                start = int(start)
+                end = int(end)
                 if start <= end:
-                    return (int(start), int(end) + 1)  # ranges are inclusive
+                    return (start, end + 1)  # ranges are inclusive
         raise ValueError(self._E_INVALID.format(self._portrange))
 
 
