@@ -17,6 +17,7 @@ from gevent import select
 #from gevent.socket import wait_read
 
 from xnet.tools import Tool
+from xnet.tools import VerbosityPrinter
 from xnet.net.ipv4 import IPRangeIterator
 
 
@@ -32,6 +33,11 @@ def _get_output_filepath(item, base):
     path = base + item
     return path
 
+
+class NetcatsException(Exception):
+    pass
+
+
 class NetcatsTool(Tool):
     '''
         NetcatsTool
@@ -105,27 +111,16 @@ class NetcatsTool(Tool):
         addr = parse_result['addr']
         sock = None
         socket_error = None
-        state = None
         result = {}
+        #
         try:
             sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
             sock.connect(addr)
-            state = 'open'
         except socket.error, e:
-            #
-            # errno numbers differ between platforms,
-            # but tests for 'closed' and 'filtered'
-            # below should work universally.
-            #
             errmsg = e.strerror.lower()
-            if 'connection refused' in errmsg:
-                state = 'closed'
-            elif 'timed out' in errmsg:
-                state = 'filtered'
-            else:
-                state = 'FIXME-' + errmsg.replace(' ', '-')
-        #except gevent.GreenletExit:
-        #    state = 'filtered'
+            self.set_error(errmsg)
+            return
+        self.vprint_stderr(1, '[connected to {0}:{1}]'.format(*addr))
 
         outfile = sys.stdout
         if self.options.split_output:
@@ -137,47 +132,50 @@ class NetcatsTool(Tool):
         done = False
         stdin = self._stdin_pipe[0]
         while not done and not sock is None:
-            rfds = [stdin, sock]
-            rlist, _, _ = select.select(rfds, [], [])
-            if stdin in rlist:
-                data = os.read(stdin, 4096)
-                if len(data) > 0:
-                    sock.send(data)
-                else:
-                    done = True
-            if sock in rlist:
-                data = sock.recv(4096)
-                if len(data) > 0:
-                    if self.options.tag_lines:
-                        tag = '\n' + self._host + ':'
-                        data = tag + data.replace('\n', '\n' + tag)
-                    #
-                    # Handle random exceptions on file-io
-                    # in a non-standard way...
-                    #
-                    while True:
-                        try:
-                            self._outfile.write(data)
-                        except IOError:
-                            pass
-                        else:
-                            break
-                    self._outfile.flush()
-                    if self.options.split_tee:
-                        sys.stdout.write(data)
-                        sys.stdout.flush()
-                    #sys.stdout.write(data)
-                    #sys.stdout.flush()
-                else:
-                    done = True
+            try:
+                rfds = [stdin, sock]
+                rlist, _, _ = select.select(rfds, [], [])
+                if stdin in rlist:
+                    data = os.read(stdin, 4096)
+                    if len(data) > 0:
+                        sock.send(data)
+                    else:
+                        done = True
+                if sock in rlist:
+                    data = sock.recv(4096)
+                    if len(data) > 0:
+                        if self.options.tag_lines:
+                            tag = '\n' + self._host + ':'
+                            data = tag + tag.join(data.split('\n'))
+                        #
+                        # Handle random exceptions on file-io
+                        # in a non-standard way...
+                        #
+                        while True:
+                            try:
+                                self._outfile.write(data)
+                            except IOError:
+                                pass
+                            else:
+                                break
+                        self._outfile.flush()
+                        if self.options.split_tee:
+                            sys.stdout.write(data)
+                            sys.stdout.flush()
+                        #sys.stdout.write(data)
+                        #sys.stdout.flush()
+                    else:
+                        done = True
+            except socket.error, e:
+                errmsg = e.strerror.lower()
+                self.set_error(errmsg)
+                try:
+                    sock.close()
+                except:
+                    pass
+                return
 
-            if 0:
-                wait_read(self._stdin_pipe[0])
-                data = self._stdin_pipe[0]
-                if len(data) > 0:
-                    sock.send(data)
-                else:
-                    done = True
+        result['addr'] = self._addr
         result['host'] = self._addr[0]
         result['port'] = self._addr[1]
         result['received'] = self._recv_buffer
@@ -186,12 +184,15 @@ class NetcatsTool(Tool):
 
     def __format__(self, line, parse_result, action_result):
         output = ''
-        if action_result and self.options.verbose:
-            output = '{0} {1} {2}'.format(
-                action_result['host'],
-                action_result['port'],
-                action_result['received'],
-            )
+        if 0:
+            if action_result and self.options.verbose:
+                output = '{0} {1} {2}'.format(
+                    action_result['host'],
+                    action_result['port'],
+                    action_result['received'],
+                )
+        host = action_result['addr']
+        self.vprint_stderr(1, '[connection closed {0}:{1}]'.format(*host))
         return output
 
     def __timeout__(self):
