@@ -12,6 +12,7 @@ import optparse
 import itertools
 import time
 import fcntl
+import subprocess
 
 import gevent
 import gevent.pool
@@ -19,6 +20,7 @@ import gevent.monkey
 from gevent.socket import wait_read
 
 import xnet.tools
+from xnet.net.ssh import SSHPopen
 
 
 gevent.monkey.patch_all(thread=False)
@@ -33,8 +35,8 @@ if DEVELOPMENT:
     from os.path import join, abspath
     sys.path.insert(0, abspath(join(os.getcwd(), '../../')))
 
-import xnet.debug
 
+DEFAULT_NR_MICROTHREADS='128'
 
 class XNetOptionParser(optparse.OptionParser):
     '''
@@ -58,8 +60,16 @@ class XNetOptionParser(optparse.OptionParser):
                         help='enable interactive pdb debugging on exceptions')
         self.add_option('', '--print-source', dest='print_source', action='store_true',
                         help='print source code of tool')
-        self.add_option('-y', '--parallelism', dest='parallelism', metavar='n',
-                        help='set parallelism (default 128)')
+        self.add_option('-x', '--nr-microthreads', dest='nr_microthreads', metavar='n',
+                        help='number of microthreads per process (default 128)')
+        self.add_option('-y', '--nr-processes', dest='nr_processes', metavar='n',
+                        help='number of processes (default 1)')
+        self.add_option('-z', '--ssh-nodes-file', dest='ssh_nodes_file', metavar='path',
+                        help='ssh workers file with user@host on each line')
+        self.add_option('', '--ssh-args', dest='ssh_args', metavar='..',
+                        help='additional arguments to the ssh command')
+        self.add_option('', '--ssh-display', dest='ssh_display', metavar='..',
+                        help='export DISPLAY=.. on ssh nodes')
         self.add_option('-r', '--read', dest='read', metavar='path',
                         help='read input from file')
         self.add_option('-w', '--wait', dest='wait', metavar='nsecs',
@@ -72,6 +82,8 @@ class XNetOptionParser(optparse.OptionParser):
                         help='print time consumed by action')
         self.add_option('-T', '--supress-timeout', dest='supress_timeout',
                         help='supress messages from tasks timing out', action='store_true')
+        self.add_option('', '--apply', dest='apply', metavar='fn',
+                        help='execute python function `fn` for each input: item = fn(item)')
         self.add_option('-f', '--format', dest='format', metavar='fmt',
                         help='specify output format, see --format-help for info')
         self.add_option('', '--format-help', dest='format_help', action='store_true',
@@ -81,7 +93,7 @@ class XNetOptionParser(optparse.OptionParser):
         self.add_option('', '--split-tee', dest='split_tee', metavar='base',
                         help='like --split-output but also print to stdout')
         self.add_option('-v', '--verbose', dest='verbose',
-                        help='verbose output', action='store_true')
+                        help='verbose output', action='count')
         self._load_local_options(local_options)
 
     def _load_local_options(self, local_options):
@@ -185,6 +197,101 @@ def repeaterator(it, count):
             count -= 1
 
 
+def do_ssh(opt, ToolImplementation, args, inputchain):
+    argv = sys.argv[:]
+    subprocesses = []
+    for ssh_node in SSHPopen.parse_ssh_nodes_file(opt.ssh_nodes_file):
+        user, host = ssh_node.split('@', 1)
+        env = {}
+        if opt.ssh_display:
+            env['DISPLAY'] = opt.ssh_display
+        ssh_args = opt.ssh_args or []
+        proc = SSHPopen(ssh_args, argv, args, user=user, host=host,
+                        stdin=subprocess.PIPE, env=env)
+        subprocesses.append(proc)
+
+    for (i, item) in enumerate(inputchain):
+        #
+        # Newlines are our input separator.
+        # Should this be argument-controlled?
+        #
+        while len(item) and item[-1] == '\n':
+            item = item[:-1]
+        item += '\n'
+        proc_index = i % len(subprocesses)
+        proc = subprocesses[proc_index]
+        proc.stdin.write(item)
+
+    for proc in subprocesses:
+        #print os.getpid(), 'SSH-CLOSE-STDIN', proc
+        proc.stdin.close()
+
+    for proc in subprocesses:
+        #print os.getpid(), 'SSH-WAIT', proc
+        proc.wait()
+    return 0
+
+
+def do_fork(opt, ToolImplementation, tool_reads_stdin, args, nr_processes, nr_microthreads, inputchain):
+    #
+    # Remove --nr-processes from sys.argv and spawn children.
+    #
+    argv = sys.argv[:]
+    if '-y' in argv:
+        i = argv.index('-y')
+        del argv[i]
+        del argv[i]
+    if '--nr-processes' in argv:
+        i = argv.index('--nr-processes')
+        del argv[i]
+        del argv[i]
+    if '--nr-processes={0}'.format(nr_processes) in argv:
+        argv.remove('--nr-processes={0}'.format(nr_processes))
+
+    #
+    # Remove all non-option args from argv before remote
+    # execution; these values will be trickled down and
+    # distributed across nodes from the parent invocation.
+    #
+    for non_option_arg in args[1:]:
+        while non_option_arg in argv:
+            i = argv.index(non_option_arg)
+            del argv[i]
+
+    if not '-' in argv:
+        argv.append('-')
+
+    subprocesses = []
+
+    #
+    # Spawn subprocesses.
+    #
+    subprocesses = []
+    for i in xrange(nr_processes):
+        proc = subprocess.Popen(argv, stdin=subprocess.PIPE)
+        subprocesses.append(proc)
+
+    for (i, item) in enumerate(inputchain):
+        #
+        # Newlines are our input separator.
+        # Should this be argument-controlled?
+        #
+        while len(item) and item[-1] == '\n':
+            item = item[:-1]
+        item += '\n'
+        proc_index = i % nr_processes
+        proc = subprocesses[proc_index]
+        proc.stdin.write(item)
+
+    for proc in subprocesses:
+        #print os.getpid(), 'CLOSE STDIN', proc
+        proc.stdin.close()
+
+    for proc in subprocesses:
+        #print os.getpid(), 'WAIT', proc
+        proc.wait()
+
+
 def run(ToolImplementation, tool_reads_stdin=False):
     try:
         return _run(ToolImplementation, tool_reads_stdin)
@@ -212,17 +319,13 @@ def _run(ToolImplementation, tool_reads_stdin):
     (opt, args) = parser.parse_args(args)
     cmdlineinputs = args[1:]
 
-    if opt.pdb:
-        xnet.debug.interactive_debugger_on_exception(True)
-
-    if opt.print_source:
-        ToolImplementation.print_source()
-        return
-
     #
     # chain input sources, handle options
     # precedence order: cmdline, -r, stdin
     #
+    # If --nr-processes=n and n > 1, then let this process
+    # produce for a number of child processes.
+    #
     _inputs = []
     if '-' in cmdlineinputs:
         if tool_reads_stdin:
@@ -236,9 +339,21 @@ def _run(ToolImplementation, tool_reads_stdin):
         _inputs.insert(0, cmdlineinputs)
     inputchain = itertools.chain(*_inputs)
 
-    _parallelism = 128
-    if not opt.parallelism is None:
-        _parallelism = int(opt.parallelism)
+    _nr_processes = 1
+    if opt.nr_processes:
+        _nr_processes = int(opt.nr_processes)
+
+    if opt.pdb:
+        import xnet.debug
+        xnet.debug.interactive_debugger_on_exception(True)
+
+    if opt.print_source:
+        ToolImplementation.print_source()
+        return
+
+    _nr_microthreads = 128
+    if not opt.nr_microthreads is None:
+        _nr_microthreads = int(opt.nr_microthreads)
 
     _wait = None
     if not opt.wait is None:
@@ -252,11 +367,6 @@ def _run(ToolImplementation, tool_reads_stdin):
     if not opt.repeat is None:
         _repeat = int(opt.repeat)
 
-    _outfile = sys.stdout
-    #
-    # main loop
-    #
-
     if opt.format_help:
         print ToolImplementation.__format_help__()
         sys.exit(0)
@@ -264,26 +374,69 @@ def _run(ToolImplementation, tool_reads_stdin):
     if opt.split_tee:
         opt.split_output = opt.split_tee
 
+    _outfile = sys.stdout
+
+    #
+    # Handle SSH-distributed execution.
+    # If SSH-dist., main process returns on do_ssh() here.
+    #
+    if opt.ssh_nodes_file:
+        #
+        # __massage__() expands wildcard IP-ranges such as 10.0.0.*.
+        # This should be done in parent in order to split the expanded
+        # set of IP:s across its children.
+        #
+        inputchain = ToolImplementation.__massage__(inputchain, opt)
+        if _repeat != 1:
+            inputchain = repeaterator(inputchain, _repeat)
+        return do_ssh(opt, ToolImplementation, args, inputchain)
+
+    #
+    # Handle multiproc.
+    # If multiproc, main process returns on do_fork() here.
+    #
+    if _nr_processes == 1:
+        pass
+    elif _nr_processes < 1:
+        errmsg = 'invalid number of processes: {0}'.format(_nr_processes)
+        sys.stderr.write(errmsg)
+        sys.exit(1)
+    else:
+        #
+        # __massage__() expands wildcard IP-ranges such as 10.0.0.*.
+        # This should be done in parent in order to split the expanded
+        # set of IP:s across its children.
+        #
+        inputchain = ToolImplementation.__massage__(inputchain, opt)
+        if _repeat != 1:
+            inputchain = repeaterator(inputchain, _repeat)
+        return do_fork(opt, ToolImplementation, tool_reads_stdin, args,
+                       _nr_processes, _nr_microthreads, inputchain)
+
     #
     # verify commandline options and do preparations
     #
     ToolImplementation.__setup__(opt)
 
-    pool = gevent.pool.Pool(_parallelism)
-    actions = []  # all spawned actions
+    #
+    # Errors if pool is too small.
+    #
+    pool = gevent.pool.Pool(_nr_microthreads)
+    #pool = gevent.pool.Pool(opt.nr_microthreads or DEFAULT_NR_MICROTHREADS)
     greenlets = []
 
     wkpool = None
     if _wait:
-        wkpool = gevent.pool.Pool(_parallelism)
+        wkpool = gevent.pool.Pool(_nr_microthreads)
 
     def waitkill(g, killtime):
+        from xnet.tools import WaitTimeout
         sleeptime = killtime - time.time()
         #print 'waitkill', sleeptime
         if sleeptime > 0:
             gevent.sleep(sleeptime)
         if not g.ready():
-            gevent.kill(g, xnet.tools.WaitTimeout)
+            gevent.kill(g, WaitTimeout)
             g.join()
 
     inputchain = ToolImplementation.__massage__(inputchain, opt)
@@ -293,15 +446,16 @@ def _run(ToolImplementation, tool_reads_stdin):
     if tool_reads_stdin:
         gevent.spawn(stdin_disperser_greenlet, pool)
 
-
     _t = 0.0
     killtime = 0
 
     if _repeat != 1:
         inputchain = repeaterator(inputchain, _repeat)
 
+    greenlet_id = 0
     for (i, line) in enumerate(inputchain):
-        action = ToolImplementation(opt, **kwargs)
+        action = ToolImplementation(opt, greenlet_id=greenlet_id, **kwargs)
+        greenlet_id += 1
         pool.wait_available()
         if _wait:
             killtime = time.time() + _wait
@@ -313,7 +467,7 @@ def _run(ToolImplementation, tool_reads_stdin):
         #
         if _wait:
             wkpool.wait_available()
-            wk = wkpool.spawn(waitkill, g, killtime)
+            wkpool.spawn(waitkill, g, killtime)
 
         #
         # handle finished actions
@@ -334,7 +488,7 @@ def _run(ToolImplementation, tool_reads_stdin):
 
     if wkpool:
         wkpool.join()
-    gevent.joinall(greenlets) #, timeout=timeout)
+    gevent.joinall(greenlets)  #, timeout=timeout)
     not_done = filter(lambda g: (not g.ready()), greenlets)
     if len(not_done) > 0 and not _wait is None:
         print 'ERROR: not_done has contents inspite of _wait and grace time'
@@ -346,6 +500,11 @@ def _run(ToolImplementation, tool_reads_stdin):
     gevent.killall(not_done, block=True)
     gevent.joinall(not_done)
 
+    #
+    # cleanups
+    #
+    ToolImplementation.__teardown__(opt)
+
     #
     # print results
     #
@@ -354,6 +513,7 @@ def _run(ToolImplementation, tool_reads_stdin):
         output_action_result(action, opt, _outfile)
         del action
 
+
 def output_action_result(a, opt, _outfile):
     outfile = _outfile
     if opt.split_output:
@@ -369,7 +529,8 @@ def output_action_result(a, opt, _outfile):
                 if not msg is None:
                     sys.stderr.write(msg)
         else:
-            print_error(a)
+            if opt.verbose >= a._error_verbosity_limit:
+                print_error(a)
     else:
         if a._action_result:
             msg = str(a)
