@@ -3,6 +3,11 @@
 #
 # Copyright(c) 2011 Krister Hedfors
 #
+# FIXME:
+#   -x -y -O is much slower than only -x -y
+#
+#   network-is-unreachable goes stdout
+#
 
 __all__ = ['TcpStateTool']
 
@@ -11,6 +16,8 @@ import socket
 
 from xnet.tools import Tool
 from xnet.tools import ToolMixin_AddrIterator
+from xnet.tools import WaitTimeout
+from xnet.debug import pdb
 
 
 class TcpStateTool(ToolMixin_AddrIterator, Tool):
@@ -55,6 +62,8 @@ class TcpStateTool(ToolMixin_AddrIterator, Tool):
             s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
             s.connect(addr)
             state = 'open'
+        except WaitTimeout:
+            raise
         except socket.error, e:
             #
             # errno numbers differ between platforms,
@@ -68,6 +77,8 @@ class TcpStateTool(ToolMixin_AddrIterator, Tool):
                 state = 'filtered'
             else:
                 state = errmsg.replace(' ', '-').replace(',', '-')
+                self.set_error(state, 1)
+                return
         #except gevent.GreenletExit:
         #    #
         #    # happens in BT, not OSX
@@ -111,6 +122,8 @@ Format variables for %s:
  port    - target port
  state   - TCP-state or error-message
  time    - time consumed by action, in ms
+ pid     - PID of executing process
+ grid    - Greenlet ID of executing process
 
 Default format:
  '{host} {port} {state}'
