@@ -17,12 +17,14 @@ __all__ = ['WebgetTool']
 #import doctest
 import sys
 import os
+import re
 import codecs
 import urlparse
 
 from xnet.tools import Tool
 from xnet.net.ipv4 import PortRangeIterator
 from xnet.net.http import UrlIterator
+from xnet.net.http import HTMLParser
 #from xnet.tools import ToolMixin_AddrIterator
 #from xnet.debug import pdb
 import xnet.packages.urllib3
@@ -74,6 +76,7 @@ class WebgetTool(Tool):
 
     MAX_REDIRECTS = 8
 
+
     def __init__(self, *args, **kw):
         super(WebgetTool, self).__init__(*args, **kw)
         if self.options.proxy:
@@ -197,6 +200,9 @@ class WebgetTool(Tool):
             _code = _response.status
             _msg = _response.reason
 
+            #
+            # Should we care about other redirect codes than these?
+            #
             if not _code in [301, 302, 303, 307]:
                 done = True
             else:
@@ -267,12 +273,15 @@ class WebgetTool(Tool):
             #_body_unicode = unicode(_body)
 
         if read_response and hasattr(self.options, 'tag_lines') and self.options.tag_lines:
-            tag = self._url + ':'
-            _body = '\n' + tag + _body.replace('\n', '\n' + tag)
+            #tag = self._url + ':'
+            #_body = '\n' + tag + _body.replace('\n', '\n' + tag)
+            tag = '\n' + self._url + ':'
+            _body = tag + tag.join(_body.split('\n'))
 
         result['host'] = _host
         result['url'] = _url
         result['response'] = _response
+        result['server'] = _response.headers.get('server', None)
         result['code'] = _code
         result['msg'] = _msg
         result['body'] = _body
@@ -283,12 +292,12 @@ class WebgetTool(Tool):
         options = self.options
         url = action_result['url']
         response = action_result['response']
+        server = action_result['server']
         code = action_result['code']
         msg = action_result['msg']
         output = ''
         #
         if options.server:
-            server = response.headers.get('server', None)
             output = '{0} {1}'.format(url, server)
         elif options.code:
             if options.verbose:
@@ -296,44 +305,14 @@ class WebgetTool(Tool):
             else:
                 output = '{0} {1}'.format(url, code)
         elif options.xpath:
-            try:
-                import lxml.etree
-            except ImportError:
-                import time
-                errmsg = '\n'
-                errmsg += '*** XNET USAGE INFORMATION ***\n'
-                errmsg += '\n'
-                errmsg += 'ImportError raised, lxml.etree might be missing.\n'
-                errmsg += 'For Debian, try:\n'
-                errmsg += '  $ sudo apt-get install python-lxml\n'
-                errmsg += 'For OSX, try:\n'
-                errmsg += '  $ sudo port install py-lxml\n'
-                print errmsg
-                time.sleep(3)
-                raise
-            htmlparser = lxml.etree.HTMLParser()
-            tree = lxml.etree.fromstring(action_result['body_unicode'], htmlparser)
+            htmlparser = HTMLParser(action_result['body_unicode'])
+            xpath_result = ''
             for xpath in self.options.xpath:
-                xpath_result = tree.xpath(xpath)
-                success = False
-                for elem in xpath_result:
-                    success = True
-                    if type(elem) is lxml.etree._ElementStringResult:
-                        elem = str(elem)
-                    if not type(elem) is str:
-                        elem = lxml.etree.tostring(elem)
-                    output += elem
-                    if len(output) and output[-1] != '\n':
-                        output += '\n'
-                if self.options.xpath_strict and not success:
-                    import tempfile
-                    (fd, name) = tempfile.mkstemp(prefix='webget-xpath-fail-')
-                    os.write(fd, action_result['body'])
-                    os.close(fd)
-                    msg = 'Error: XPath not matched: {0}'.format(xpath)
-                    msg += ', body saved to {0}'.format(name)
-                    sys.exit(msg)
-                output += self.options.xpath_delimiter or ' '
+                if len(xpath_result):
+                    xpath_result += '\n'
+                xpath_result += htmlparser.xpath(xpath,
+                                                 match_delimiter=self.options.xpath_delimiter)
+            output = xpath_result
         else:
             output = action_result['body']
         return output
@@ -345,10 +324,13 @@ Format variables for %s:
 
  host     - target host
  url      - target URL
+ server   - Server header value
  code     - status code
  msg      - status message
  body     - response body
  time     - time consumed by action, in ms
+ pid      - PID of executing process
+ grid     - Greenlet ID of executing process
 
 Default format depend on commandline arguments.
         ''' % cls.__toolname__
